/*
 ************************************************************************
 *
 *			   				Root Module
 * Note on animation.
 * We don't use null events to do animation: all animation is done in the
 * drawing function itself. To activate animation, we invalidate regions
 * covering both 2d and 3d views, and set is_flying to TRUE. Invalidating regions
 * guarantees that visRgn of our on-screen window is set to include both views,
 * and it makes sure that the next event would be an Update event. This event
 * is eventually relayed to our 2d_view's and 3d_view's drawing functions.
 * When is_flying is set to TRUE, these functions not only draw a window, but
 * perform a few cycles of animation as well. Before they return, they invalidate
 * the regions again, thus making sure that a new Update event would be generated,
 * and they'll be called again.
 *
 ************************************************************************
 */

#include "fractal_map.h"
#include "ImageViews.h"

#include <Debug.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>


/*--------------------------------------------------------*/
	/*
	 The only job of this view is to blit the offscreen
	 bits into the onscreen view.
	*/


class TBitmapView : public BView {
	B_DECLARE_CLASS_INFO(BView);

protected:
	BBitmap&	the_bitmap;

public:
		TBitmapView(BRect frame, BBitmap& _bitmap)
		: BView(frame, "", B_FOLLOW_NONE, B_WILL_DRAW),
		  the_bitmap(_bitmap) {}

virtual	void	Draw(BRect updateRect)
			{ DrawBitmap(&the_bitmap, updateRect, updateRect); }
};

B_DEFINE_CLASS_INFO(TBitmapView, BView);



/*--------------------------------------------------------*/
	/*
	 Blit the offscreen bits into the onscreen view
	 plus draw the plane
	*/



class TBitmapViewAndPlane : public TBitmapView {
	B_DECLARE_CLASS_INFO(TBitmapView);

	const ViewerPosition& viewer_pos;
	
public:
	TBitmapViewAndPlane(BRect frame, BBitmap& _bitmap,
					const ViewerPosition& _viewer_pos)
		: TBitmapView(frame,_bitmap),
		  viewer_pos(_viewer_pos) {}
    virtual void Draw(BRect update);
};

B_DEFINE_CLASS_INFO(TBitmapViewAndPlane, TBitmapView);


void TBitmapViewAndPlane::Draw(BRect update)
{
	DrawBitmap(&the_bitmap, update, update);

	SetPenSize(1.0);
    const rgb_color color = {150,30,30,10};
	BPoint pt(viewer_pos.xe,viewer_pos.ye);
	BeginLineArray(2);
	AddLine(pt-BPoint(viewer_pos.gx>>13,viewer_pos.gy>>13),
	        pt+BPoint(viewer_pos.gx>>11,viewer_pos.gy>>11),color);
	AddLine(pt-BPoint(viewer_pos.gy>>13,viewer_pos.gx>>13),
	        pt+BPoint(viewer_pos.gy>>12,viewer_pos.gx>>12),color);
	EndLineArray();
}

/*--------------------------------------------------------*/

class Clouds2DWin : public BWindow {
	B_DECLARE_CLASS_INFO(BWindow);

  ImageView image_2d_view;
  const ViewerPosition& viewer_pos;
  TBitmapViewAndPlane	*fMainView;

  virtual bool QuitRequested(void)
  {			// Tell the app to go ahead and quit.
	be_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
  }

public:
  Clouds2DWin(const IMAGE& image, const ViewerPosition& _viewer_pos);

virtual	void		MessageReceived(BMessage *msg);
		void		Update(BRect rect);
		void		Update(void)
					{ Update(fMainView->Bounds()); }
//		void		DrawOffScreen(BView *v);
};

B_DEFINE_CLASS_INFO(Clouds2DWin, BWindow);


Clouds2DWin::Clouds2DWin(const IMAGE& image,const ViewerPosition& _viewer_pos)
	: BWindow(ScreenRect(image)+=100, "view from above", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE),
	  image_2d_view(image), viewer_pos(_viewer_pos)
{
	Lock();

  ScreenRect rect(image);
  rect.offset_to_origin();

	/*
	 The drawing will take place in the view fOffView that
	 will be added to the offscreen bitmap fBitmap. In 
	 this way we'll do the drawing offscreen and then just
	 blit the result to the screen.
	*/
//	fBitmap = new BBitmap(rect, B_COLOR_8_BIT, TRUE);
//	fOffView = new BView(rect, "", 0, B_WILL_DRAW);

//	fBitmap->Lock();
//	fBitmap->AddChild(fOffView);
//	DrawOffScreen(fOffView);	// draw the initial setup
//	fBitmap->Unlock();

	/*
	 This view will be added to the visible window. It's
	 only role is to blit the offscreen bitmap to the
	 visible window.
	*/
	fMainView = new TBitmapViewAndPlane(rect, image_2d_view,viewer_pos);
	AddChild(fMainView);

	Unlock();
 
  Show();
}

class Clouds3DWin : public BWindow {
	B_DECLARE_CLASS_INFO(BWindow);

  ProjectionParameters projection_parms;


  int	time_started;					// Time (ticks) started flying the last time
  int	frame_counter;
  
  ThreeDView image_3d_view;
  const ViewerPosition& viewer_pos;
  TBitmapView	*fMainView;

  virtual bool QuitRequested(void)
  {			// Tell the app to go ahead and quit.
	be_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
  }


public:
  Clouds3DWin(const IMAGE& image,const ViewerPosition& viewer_pos);


virtual	void		MessageReceived(BMessage *msg);
		void		Update(BRect rect) { Update(); }
		void		Update();
//		void		DrawOffScreen(BView *v);
};

B_DEFINE_CLASS_INFO(Clouds3DWin, BWindow);


Clouds3DWin::Clouds3DWin(const IMAGE& image,const ViewerPosition& _viewer_pos)
	: BWindow(ScreenRect(image)+=200, "view from the cockpit", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE),
	  image_3d_view(image,_viewer_pos,projection_parms),
	  viewer_pos(_viewer_pos)
{
	Lock();

  ScreenRect rect(image);
  rect.offset_to_origin();

    /*
	 This view will be added to the visible window. It's
	 only role is to blit the offscreen bitmap to the
	 visible window.
	*/
	fMainView = new TBitmapView(rect, image_3d_view);
	AddChild(fMainView);

	//fSkipFrameMoved = fSkipDraw = 0;


  projection_parms.ze = 350;//eyepoint_elevation;
  projection_parms.beta = 83; //beta;
  projection_parms.Ou   = 128;//Ou;
  projection_parms.Ov   = 270;//Ov;
  projection_parms.z_cutoff = 0;//z_cutoff;
  projection_parms.elevation_factor = 700;//elevation_factor;
  image_3d_view.re_project();
	Unlock();
 
  Show();
}


#if 0
								// Commencing and terminating the flight
void Clouds2DWin::start_flight(void)
{
  is_flying = TRUE;
//  ModelessDialog::ControlItem(*this,item_Go).set_title("\pStop");
  frame_counter = 0;
  time_started = TickCount();
}

void Clouds2DWin::stop_flight(void)
{
  is_flying = FALSE;
//  ModelessDialog::ControlItem(*this,item_Go).set_title("\pGo");
  const int time_elapsed = TickCount() - time_started;	// in Ticks
//  NumberItem(*this,item_fps).draw(
//  		time_elapsed == 0 ? 0 : (60*frame_counter)/time_elapsed );
}
#endif



/*--------------------------------------------------------*/

void Clouds2DWin::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case BB_DRAW:
			{
			/*
			 Cheezy flow control. If there are 3 or more BB_DRAW events
			 already in the queue, then skip this one, and, at most the next
			 message.
			*/
			//if ((fSkipDraw < 2) && MessageQueue()->FindMessage(BB_DRAW, 2)) {
			//	fSkipDraw++;
			//	break;
			//}
			//fSkipDraw = 0;

			// The plane is telling the window to redraw
			Update(msg->FindRect("update"));
			break;
			}
		case BB_HIT_HOLE:
			{
			#if 0
			/*
			 The ball is telling us that it just hit the
			 hole. So it should get sent to the partner.
			*/
			BMessage	*m = new BMessage(BB_BALL);
			m->AddPoint("speed", msg->FindPoint("speed"));
			m->AddFloat("rel_loc", msg->FindFloat("rel_loc"));


			// delete the ball from this window
			fBall->Quit();
			fBall = NULL;

			Update();
			#endif
			break;
			}
		case BB_STOP:
			{
			// Stop the ball in mid-flight
			//if (fBall) {
			//	fBall->SetEnabled(!fBall->IsEnabled());

			//	BMenuItem *item = cast_as(msg->FindObject("source"), BMenuItem);
			//	item->SetMarked(!item->IsMarked());
			//}
			break;
			}
		default:
			inherited::MessageReceived(msg);
			break;
	}
}

void Clouds3DWin::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case BB_DRAW:
			{
			// The plane is telling the window to redraw
			Update(msg->FindRect("update"));
			break;
			}
		case BB_HIT_HOLE:
			break;

		case BB_STOP:
			break;

		default:
			inherited::MessageReceived(msg);
			break;
	}
}

#if 0
/*--------------------------------------------------------*/

bool TWindow::QuitRequested()
{
	/*
	 The window was asked to quit/close. Send a message
	 to our partner, giving him the ball if we've
	 currently got it.
	*/
	BMessage	*m = new BMessage(BB_GOODBYE);
	if (fBall) {
		fBall->Quit();
		fBall = NULL;
		m->AddBool("ball", TRUE);
	}
	my_app->SendToPartner(m);

	// Tell the app to go ahead and quit.
	my_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
}

/*--------------------------------------------------------*/

#endif

/*--------------------------------------------------------*/

void Clouds2DWin::Update(BRect rect)
{
Lock();
	// redraw window, only blitting the given rect.
//	DrawOffScreen(fOffView);
	fMainView->Draw(rect);
Unlock();
}


/*--------------------------------------------------------*/

void Clouds3DWin::Update(void)
{
Lock();
	// redraw window, blitting the entire window
    image_3d_view.re_project();
    fMainView->Draw(fMainView->Bounds());
Unlock();
}

/*--------------------------------------------------------*/

#if 0
void Clouds2DWin::DrawOffScreen(BView *v)
{
	/*
	 Do the actual drawing of the 'room'. The walls will get
	 drawn (with a gap if there is one) and the ball will
	 be placed in the proper position.
	*/
	fBitmap->Lock();

	BRect b = v->Bounds();
	
	v->SetPenSize(2.0);
	v->SetHighColor(100,100,0);
	v->FillRect(b, B_SOLID_LOW);

	b.bottom -= 1.0;
	b.right -= 1.0;

	// draw the appropriate border - with or without a 'gap'
	if (fGapStart == -1.0)
		v->StrokeRect(b);
	else
		DrawFrame(v, b, fGapStart, fGapEnd);

	if (fBall)
		fBall->Draw(v);

	v->Sync();			// flush all drawing to the bitmap

	fBitmap->Unlock();
}
#endif

/*--------------------------------------------------------*/
/*--------------------------------------------------------*/
/*--------------------------------------------------------*/


/*--------------------------------------------------------*/
/*--------------------------------------------------------*/
/*--------------------------------------------------------*/


/*--------------------------------------------------------*/
										// Making elevation maps to fly around to
										
class MakeTestImage : public LazyImage	// this one makes a couple of pyramids
{
  void fill_in(IMAGE& im) const
  {
	const int ncols = im.q_ncols();
	for(register int j=-20; j<20; j++)
	  for(register int i=-5; i<5; i++)
	    im(64+i,j+ncols/2) = 128 - 5*abs(i+j),
	    im(64+40+i,j+ncols/2) = 180 - 7*abs(i+j);
  }
public:
  MakeTestImage(void) : LazyImage(256,256,8) {}
};

class GaussClouds : public FractalMap
{
  public:
    GaussClouds(const card order)
      : FractalMap(order) {}
				// Well-known Gaussian random number generator
				// Note rand() returns a number [0,2^15-1],
				// that is, within [0,1] with a 15-bit implicit
				// fraction
    inline int get_noise(const card scale) const {
      long sum = 0;
      for(register int i=0; i<12; i++)
	     sum += rand();			// keep the result within 
      return (scale * (sum-(6<<15)))>>17; }	// [-scale/2,scale/2]
};


/*--------------------------------------------------------*/

/*
 The Application object for the BeBounce application.
 Passing my signature to the system
*/

class TVenus : public BApplication {
	B_DECLARE_CLASS_INFO(BApplication);

enum {MY_SIG = 'venu'};
		Clouds2DWin		*window_2D;
		Clouds3DWin		*window_3D;
		MyPlane * my_plane;
		IMAGE image;
  ViewerPosition viewer_pos;						// Where the viewer is at
public:
					TVenus();
virtual				~TVenus();
};

// ??? should I use these CLASS_INFO macros or not ???
B_DEFINE_CLASS_INFO(TVenus, BApplication);

TVenus::TVenus()
	: BApplication(MY_SIG), image(GaussClouds(8)),
	  viewer_pos(110,128, 1<<14,0)
{

#if 0
  IMAGE image = GaussClouds(8);					// Get gaussian clouds and smooth them a bit
  FilterIt(image).conv(conv_kernel(1,1,1,CommonDenom(3))).clip_to_intensity_range();
  IMAGE image = MakeTestImage();
#endif

  window_2D = new Clouds2DWin(image,viewer_pos);
  window_3D = new Clouds3DWin(image,viewer_pos);

  my_plane = new MyPlane(*window_2D, *window_3D,viewer_pos);

#if 0
	/*
	 A little bit of menu code. Let's add a 'Stop Ball'
	 menu item to the application menu (found in the
	 Dock window)
	*/
	BMenuItem	*item;
	BPopUpMenu	*menu = MainMenu();
	item = new BMenuItem("Stop Ball",
					new BMessage(BB_STOP), 'K');
	item->SetTarget(fWindow);
	menu->AddItem(item, 1);
#endif
}

/*--------------------------------------------------------*/

TVenus::~TVenus()
{
    //my_plane->Quit();
	window_2D = NULL;
	window_3D = NULL;
}


main(int argc, char* argv[])
{
	/*
	 To randomize the movement of the bouncing ball set
	 the random seed to the system time
	*/
	srand((long) system_time());


	// make the new application object and start it running
	TVenus().Run();
	return 0;
}

